from .models import Session
from abc import ABC, abstractmethod
from typing import List
from .types import StoreValue, StoreContainer


class StateStore(ABC):
    def __init__(self):
        """
        Create the container.
        """

        self.container: StoreContainer = {}

    @abstractmethod
    def all(self, session: Session) -> List[StoreValue]:
        """
        Get all entries for a session in the container.
        """

        pass  # pragma: no cover

    @abstractmethod
    def append(self, session: Session, value: StoreValue) -> None:
        """
        Append an entry for a session to the container.
        """

        pass  # pragma: no cover

    @abstractmethod
    def reset(self, session: Session) -> None:
        """
        Reset the container for a session back to empty.
        """

        pass  # pragma: no cover


class TimeMemoryStore(StateStore):
    def all(self, session: Session) -> List[StoreValue]:
        domain = session.domain
        if domain not in self.container:
            # Trigger creation of dict for shop
            self.reset(session)
        return self.container[domain]

    def append(self, session: Session, value: StoreValue) -> None:
        domain = session.domain
        if domain not in self.container:
            # Trigger creation of dict for shop
            self.reset(session)
        self.container[domain].append(value)

    def reset(self, session: Session) -> None:
        self.container[session.domain] = []


class CostMemoryStore(StateStore):
    def all(self, session: Session) -> List[StoreValue]:
        domain = session.domain
        if domain not in self.container:
            # Trigger creation of dict for shop
            self.reset(session)
        return self.container[domain]

    def append(self, session: Session, value: StoreValue) -> None:
        domain = session.domain
        if domain not in self.container:
            # Trigger creation of dict for shop
            self.reset(session)
        self.container[domain].append(value)

    def reset(self, session: Session) -> None:
        self.container[session.domain] = []



from typing import Optional
from httpx._models import Response
from http import HTTPStatus
from .types import ParsedBody, ParsedError


class Session:
    def __init__(self, domain: str = None, key: str = None, password: str = None, secret: str = None):
        self.domain = domain
        self.key = key
        self.password = password
        self.secret = secret

    @property
    def base_url(self) -> str:
        return f"https://{self.domain}"


class RestLink:
    def __init__(self, next: Optional[str], prev: Optional[str]):
        self.next = next
        self.prev = prev


class ApiResult:
    def __init__(
        self,
        response: Response,
        status: HTTPStatus,
        body: ParsedBody,
        errors: ParsedError,
        retries: int = 0,
    ):
        self.response = response
        self.status = status,
        self.body = body
        self.errors = errors
        self.retries = retries


class RestResult(ApiResult):
    def __init__(self, link: RestLink, **kwargs):
        super().__init__(**kwargs)
        self.link = link



from .__version__ import VERSION
from .options import Options
from .clients import Client, AsyncClient, ApiCommon
from .models import ApiResult, RestResult, Session
from .store import CostMemoryStore, TimeMemoryStore, StateStore
from .deferrer import Deferrer, SleepDeferrer



from typing import Union
import hashlib
import hmac
import base64


# Encoding format
e = "utf-8"


def create_hmac(
    data: Union[dict, str],
    raw: bool = False,
    build_query: bool = False,
    build_query_with_join: bool = False,
    encode: bool = False,
    secret: str = None
) -> str:
    """
    Create an HMAC string based on inputted values.
    """

    if build_query:
        # Query building is required, sort the keys alphabetically
        sorted_keys = sorted(data.keys())
        query_string = []
        for key in sorted_keys:
            value = data[key]
            # Join arrays together by ","
            query_value = ",".join(value) if isinstance(value, (tuple, list)) else value
            query_string.append(f"{key}={query_value}")
        # Optionally join result by "&"
        join_key = "&" if build_query_with_join else ""
        data = join_key.join(query_string).encode(e)

    # Generate the HMAC value
    hmac_local = hmac.new(secret.encode(e), data, hashlib.sha256)
    if encode:
        # For webhooks
        return base64.b64encode(hmac_local.digest())
    # For 0Auth and proxy
    return hmac_local.hexdigest().encode(e)


def hmac_verify(source: str, secret: str, params: dict, hmac_header: str = None) -> bool:
    """
    Verify if the HMAC is correct.
    """

    if source == "standard":
        # Standard 0Auth/URL method
        hmac_param = params["hmac"].encode(e)
        params.pop("hmac", None)
        kwargs = {
            "data": params,
            "build_query": True,
            "build_query_with_join": True,
        }
    elif source == "proxy":
        # Proxy app request method
        hmac_param = params["signature"].encode(e)
        params.pop("signature", None)
        kwargs = {
            "data": params,
            "build_query": True,
            "build_query_with_join": False,
        }
    elif source == "webhook":
        # Webhook data method
        hmac_param = hmac_header.encode(e)
        kwargs = {
            "data": params.encode(e),
            "raw": True,
            "encode": True,
        }

    # Create the HMAC and compare to what was supplied
    hmac_local = create_hmac(**kwargs, secret=secret)
    return hmac.compare_digest(hmac_local, hmac_param)



import time
import asyncio
from abc import ABCMeta, abstractmethod
from .types import SleepTime


class Deferrer(metaclass=ABCMeta):
    def current_time(self) -> int:
        """
        Get the current time in ms.
        """

        return int(round(time.time() * 1000))

    @abstractmethod
    def sleep(self, length: SleepTime) -> None:
        """
        Sleep (sync) for X ms.
        """

        pass  # pragma: no cover

    @abstractmethod
    def asleep(self, length: SleepTime) -> None:
        """
        Sleep (async) for X ms.
        """

        pass  # pragma: no cover


class SleepDeferrer(Deferrer):
    def sleep(self, length: SleepTime) -> None:
        time.sleep(length / 1000.0)

    async def asleep(self, length: SleepTime) -> None:
        asyncio.sleep(length / 1000.0)



# Valid API version in URL path: YYYY-MM or unstable
VERSION_PATTERN = r"([0-9]{4}-[0-9]{2})|unstable"
# /oauth/authorize, /oauth/access_token do not require authentication
NOT_AUTHABLE_PATTERN = r"\/oauth\/(authorize|access_token)"
# /oauth/access_scopes does not require versioned API path
NOT_VERSIONABLE_PATTERN = r"\/(oauth\/access_scopes)"
# For extracting next/prev link header on REST API calls
LINK_PATTERN = r"<.*page_info=([a-zA-Z0-9\-_]+).*>; rel=\"(next|previous)\""
# Header supplied by Shopify when rate limit is hit
RETRY_HEADER = "retry-after"
# Header supplied by Shopify for REST API calls, used by LINK_PATTERN
LINK_HEADER = "link"
# Header to send for public API calls
ACCESS_TOKEN_HEADER = "x-shopify-access-token"
# Default API version
DEFAULT_VERSION = "2020-04"
# Default API mode
DEFAULT_MODE = "public"
# Alternate API mode
ALT_MODE = "private"
# One second in milliseconds
ONE_SECOND = 1000
# REST API type
REST = "rest"
# GraphQL API type
GRAPHQL = "graphql"



from typing import Union, List, Dict, Optional
from httpx._types import RequestData, QueryParamTypes

# Joins HTTPX's query or post data
UnionRequestData = Union[QueryParamTypes, RequestData]
# Value for cost/time storage
StoreValue = Union[int, float]
# Container dict setup for cost/time storage
StoreContainer = Dict[str, List[StoreValue]]
# Parsed JSON body from response
ParsedBody = Optional[dict]
# Parsed error body from response
ParsedError = Optional[Union[dict, Exception]]
# Time to sleep for deferrer
SleepTime = Union[float, int]



from os.path import abspath, dirname, join
from setuptools import setup

NAME = "basic_shopify_api"
exec(open("basic_shopify_api/__version__.py").read())
DESCRIPTION = "A sync/async REST and GraphQL client for Shopify API using HTTPX"
with open(join(dirname(abspath(__file__)), "README.md")) as readme:
    README = readme.read()

setup(
    name=NAME,
    version=VERSION,
    description="A sync/async REST and GraphQL client for Shopify API using HTTPX",
    long_description_content_type="text/markdown",
    long_description=README,
    author="osiset",
    author_email="tyler@osiset.com",
    url="https://github.com/osiset/basic_shopify_api",
    packages=["basic_shopify_api"],
    license="MIT License",
    install_requires=[
        "httpx>=0.13"
    ],
    test_suite="tests",
    tests_require=[
        "mock>=1.0.1",
    ],
    platforms="Any",
    python_requires=">=3.6",
    zip_safe=False,
    include_package_data=True,
)



from http import HTTPStatus
from .store import TimeMemoryStore, CostMemoryStore
from .deferrer import SleepDeferrer
from .constants import DEFAULT_VERSION, DEFAULT_MODE, ALT_MODE, VERSION_PATTERN
import re


class Options(object):
    def __init__(self):
        # Number of retries that should happen if failed
        self.max_retries = 2
        # Retry if status is 422, 5xx
        self.retry_on_status = [
            HTTPStatus.TOO_MANY_REQUESTS.value,
            HTTPStatus.BAD_GATEWAY.value,
            HTTPStatus.SERVICE_UNAVAILABLE.value,
            HTTPStatus.GATEWAY_TIMEOUT.value,
        ]
        # Always send these headers with every request
        self.headers = {
            "Content-Type": "application/json",
            "Accept": "application/json"
        }
        # Time storage implementation (REST & GraphQL)
        self.time_store = TimeMemoryStore()
        # Cost storage implementation (GraphQL)
        self.cost_store = CostMemoryStore()
        # Deferrer implementation for getting current time and sleeping
        self.deferrer = SleepDeferrer()
        # Number of calls per second for REST... 2 for regulatr, 4 for plus
        self.rest_limit = 2
        # Number of cost points allowed per second for GraphQL... 50 for regular, 100 for plus
        self.graphql_limit = 50
        # Methods to run before firing REST API calls
        self.rest_pre_actions = []
        # Methods to run after firing REST API calls
        self.rest_post_actions = []
        # Methods to run before firing GraphQL API calls
        self.graphql_pre_actions = []
        # Methods to run after firing GraphQL API calls
        self.graphql_post_actions = []
        # Version to use for API calls
        self._version = DEFAULT_VERSION
        # Mode to use... public or private
        self._mode = DEFAULT_MODE

    @property
    def version(self) -> str:
        return self._version

    @version.setter
    def version(self, value: str) -> None:
        if not bool(re.match(VERSION_PATTERN, value)):
            raise ValueError(f"Version: {value} does not match YYYY-MM format or unstable")
        self._version = value

    @property
    def mode(self) -> str:
        return self._mode

    @mode.setter
    def mode(self, value: str) -> None:
        if value != DEFAULT_MODE and value != ALT_MODE:
            raise ValueError(f"Type must be either {DEFAULT_MODE} or {ALT_MODE}")
        self._mode = value

    @property
    def is_public(self) -> bool:
        return self.mode == DEFAULT_MODE

    @property
    def is_private(self) -> bool:
        return not self.is_public
